/*
 * vm-restart - simple daemon to detect VMware guest OS restart
 *
 * Copyright (c) 2009, Paul Chakravarti paul.chakravarti@gmail.com   
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *      Uses VMWare backdoor code from VM Back - VMware Command Line Tools
 *      http://chitchat.at.infoseek.co.jp/vmware/vmtools.html
 *
 *      Copyright (c) 2006 Ken Kato
 */

#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <unistd.h>
#include <math.h>
#include <time.h>

#include "vm.h"

void error_syslog(char *message,...) {
    va_list ap;
    va_start(ap,message);
    vsyslog(LOG_ERR,message,ap);
    va_end(ap);
    exit(EXIT_FAILURE);
}

void error_stderr(char *message,...) {
    va_list ap;
    va_start(ap,message);
    verrx(EXIT_FAILURE,message,ap);
    va_end(ap);
}

void info_syslog(char *message,...) {
    va_list ap;
    va_start(ap,message);
    vsyslog(LOG_INFO,message,ap);
    va_end(ap);
}

void info_stderr(char *message,...) {
    va_list ap;
    va_start(ap,message);
    vwarnx(message,ap);
    va_end(ap);
}

void debug_null(char *message,...) {
}

void (*error)(char *message,...) = error_stderr;
void (*info)(char *message,...) = info_stderr;
void (*debug)(char *message,...) = debug_null;

double getGuestTime(){
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME,&ts);
    return (double)ts.tv_sec + ((double)(ts.tv_nsec/1000))/1000000;
}

double getHostTime() {
    uint32_t time_s,time_us;
    if (VMGetGmTime(&time_s,&time_us,NULL,NULL) != VMTOOL_SUCCESS) {
        error("VmGetGmTime Error");
    }
    return (double)time_s + ((double)time_us)/1000000;
}

void setTime() {
    uint32_t time_s,time_us;
    struct timespec ts;
    if (VMGetGmTime(&time_s,&time_us,NULL,NULL) != VMTOOL_SUCCESS) {
        error("VmGetGmTime Error");
    }
    ts.tv_sec = (time_t) time_s;
    ts.tv_nsec = (long) time_us * 1000L;
    clock_settime(CLOCK_REALTIME,&ts);
}

void syncTime() {
    double guest_delta;
    info("Synchronising with host time");
    do {
        setTime();
        sleep(1);
        guest_delta = fabs(getHostTime() - getGuestTime());
        debug("Difference = %.3f",guest_delta);
    } while (guest_delta > 0.1); 
}

int main(int argc, char **argv) {

    double host,last_host,guest,guest_delta,host_delta;
    char *prog_name;
    extern char *optarg;
    extern int optind;
    int ch;
    int verbose = 0; 
    int daemonise = 1;
    int interval = 1;
    int saved_pid = 0;
    double restart_threshold = 3.0;
    double limit_threshold = 0.0;
    int sync_restart = 0;
    int sync_once = 0;
    int sync_limit = 0;
    char *restart_cmd = NULL;
    FILE *pid_fp;
    char *pid_file = "/var/run/vm-restart.pid";
    
    if (prog_name = rindex(*argv,'/'))
        prog_name++;
    else
        prog_name = *argv;

    while ((ch = getopt(argc,argv,"Di:r:se:p:t:yvh")) != -1) {
        switch (ch) {
            case 'D':
                daemonise = 0;
                break;
            case 'i':
                interval = atoi(optarg);
                restart_threshold = (float)interval + 2.0;
                break;
            case 'r':
                restart_threshold = atof(optarg);
                break;
            case 's':
                sync_restart = 1;
                break;
            case 'e':
                restart_cmd = optarg;
                break;
            case 'p':
                pid_file = optarg;
                break;
            case 't':
                sync_limit = 1;
                limit_threshold = atof(optarg);
                break;
            case 'y':
                sync_once = 1;
                break;
            case 'v':
                verbose = 1;
                debug = info;
                break;
            case 'h':
                printf("Usage: %s [-Dvsy] [-i <interval>] [-t <threshold>] [-p <pid_file>] [-e <cmd>]\n",prog_name);  
                printf("       -D                Debug - do not daemonise\n");
                printf("       -i <interval>     Poll interval (secs) [1]\n");
                printf("       -r <threshold>    Restart threshold (secs) [3.0]\n");
                printf("       -s                Sync time to host on restart\n");
                printf("       -e <restart_cmd>  Shell command to exceute on restart\n");
                printf("       -p <pid_file>     PID file (%s)\n",pid_file);
                printf("       -t <limit>        Sync time with host if limit exceeded (forward only)\n");
                printf("       -y                Sync time once with host and exit\n");
                printf("       -v                Verbose\n");
                printf("       -h                Print help\n");
                exit(EXIT_FAILURE);
                break;
        }
    }


    if (VMCheckVirtual() != VMTOOL_SUCCESS) {
        error("Error - not running in VM");
    }

    if (sync_once || sync_restart || sync_limit) {
#if defined(__NetBSD__)
        if ((geteuid() != 0) && (access("/dev/clockctl",W_OK) != 0)) 
#else
        if (geteuid() != 0) 
#endif
            error("ERROR: Must be root to set system time");
    }

    if (sync_once) {
        syncTime();
        exit(0);
    }

    if (daemonise) {

        if ((pid_fp = fopen(pid_file,"r")) != NULL) {
            fscanf(pid_fp,"%d",&saved_pid);
            if (saved_pid != 0) {
                if (kill(saved_pid,0) ==0) {
                    error("already running - pid: %d",saved_pid);
                }
            }
            fclose(pid_fp);
        }
        if ((pid_fp = fopen(pid_file,"w")) == NULL) {
            error("unable to open pid file %s: %s", pid_file, strerror(errno));
        }

        if (daemon(0,0) == -1) {
            error("%s: daemon(3) failed: %s",prog_name,strerror(errno));
        }

        openlog(prog_name,LOG_PID,LOG_DAEMON);

        error = error_syslog;
        info = info_syslog;
        if (verbose) {
            debug = info_syslog;
        }

        fprintf(pid_fp,"%d\n",getpid());
        fclose(pid_fp);
    }

    info("interval=%d, restart_threshold=%.3f, sync_restart=%d, restart_cmd=\"%s\", sync_limit=%d, limit_threshold=%.3f" 
                ,interval,restart_threshold,sync_restart,restart_cmd,sync_limit,limit_threshold);

    last_host = getHostTime();

    while(1) {
        sleep(interval);
        host = getHostTime();
        guest = getGuestTime();
        host_delta = host - last_host;
        guest_delta = host - guest;
        debug("Host=%.3f, Guest=%.3f, Host_Delta=%.3f (%.3f), Guest_Delta=%.3f (%.3f)",
                host,guest,host_delta,restart_threshold,guest_delta,limit_threshold); 
        if (host_delta > restart_threshold) {
            info("Restart Threshold Exceeded (%.1f)",host_delta);
            if (sync_restart) {
                syncTime();
            }
            if (restart_cmd) {
                info("Executing restart_cmd: %s", restart_cmd);
                system(restart_cmd);
            }
            host = getHostTime();
        } else if (sync_limit && (guest_delta > limit_threshold)) {
            syncTime();
            host = getHostTime();
        }
        last_host = host;
    }

    return EXIT_SUCCESS;
}
